# -*- coding: utf-8 -*-
"""
plaid - plaid looks at integrated data
F.H. Gjørup 2025-2026
Aarhus University, Denmark
MAX IV Laboratory, Lund University, Sweden

This module provides functions to interact with NeXus HDF5 files

"""
import h5py as h5

def get_nx_group(gr, name, nxclass=None):
    """Get a generic nexus group with a specific name or nxclass from a group."""
    if gr is None:
        return None
    if name in gr:
        return gr[name]
    if nxclass is not None:
        for key in gr.keys():
            if "NX_class" in gr[key].attrs and gr[key].attrs["NX_class"] == nxclass:
                return gr[key]

def get_h5_dset_path(gr,name):
    """Utility function to get the full path of a dataset in an HDF5 file group."""
    if gr is None:
        return None
    if name in gr:
        return gr[name].name
    return None

def get_nx_entry(f,definition=None,allow_subentry=True):
    """
    Get the entry nexus group from a nexus hdf5 instance.
    If a definition is given (e.g. 'NXazint1d'), return the entry or subentry
    with the matching definition attribute. If no matching definition is found,
    return None. If allow_subentry is True, also search for subentries.
    If no definition is given, return the main entry or subentry (if allow_subentry is True).
    """

    def matching_definition(gr,definition):
        """Check if the group has a matching definition attribute. Case insensitive."""
        if 'definition' in gr:
            entry_definition = gr['definition'][()]
            if isinstance(entry_definition, bytes):
                entry_definition = entry_definition.decode('utf-8')
            if entry_definition.lower() == definition.lower():
                return True
        return False
    
    allowed_classes = ['NXentry', 'NXsubentry'] if allow_subentry else ['NXentry']
    if isinstance(f, h5.Group):
        # if f is already a group, check if it is an entry or subentry
        if 'NX_class' in f.attrs and f.attrs['NX_class'] in allowed_classes:
            # if a definition is given, check if it matches
            if definition is None or matching_definition(f,definition):
                return f
        # if the group is not an entry or subentry with the correct definition
        # or no definition was given, start from the root of the file
        f = f.file
    # if f is a file, get the entry group
    entry = get_nx_group(f, 'entry', 'NXentry')
    if entry is None:
        return None
    if definition is None or matching_definition(entry,definition):
        return entry
    if not allow_subentry:
        return None
    # check the nxentry for all nxsubentries and return the one with the correct definition
    for key in entry.keys():
        if "NX_class" in entry[key].attrs and entry[key].attrs["NX_class"] == "NXsubentry":
            if matching_definition(entry[key],definition):
                return entry[key]
    # if no valid definition was found, use the group names as a fallback
    for key in entry.keys():
        if "NX_class" in entry[key].attrs and entry[key].attrs["NX_class"] == "NXsubentry":
            if key.lower() in definition.lower():
                return entry[key]
    # else return None
    return None

def get_nx_monitor(gr):
    """Get the monitor nexus group from a nexus hdf5 file."""
    gr = get_nx_entry(gr)
    return get_nx_group(gr, 'monitor', 'NXmonitor')

def get_nx_sample(gr):
    """Get the sample nexus group from a nexus hdf5 file."""
    gr = get_nx_entry(gr)
    return get_nx_group(gr, 'sample', 'NXsample')

def get_nx_transformations(gr):
    """Get the transformations nexus group from a nexus hdf5 file."""
    gr = get_nx_sample(gr)
    return get_nx_group(gr, 'transformations', 'NXtransformations')

def get_nx_instrument(gr):
    """Get the instrument nexus group from a nexus hdf5 file."""
    if gr is None:
        return None
    # check if the group is already an instrument
    if 'NX_class' in gr.attrs and gr.attrs['NX_class'] == 'NXinstrument':
        return gr
    gr = get_nx_entry(gr)
    return get_nx_group(gr, 'instrument', 'NXinstrument')

def get_nx_monochromator(gr):
    """Get the nxmonochromator group from a nexus hdf5 file."""
    gr = get_nx_instrument(gr)
    return get_nx_group(gr, 'monochromator', 'NXmonochromator')

def get_nx_source(gr):
    """Get the source nexus group from a nexus hdf5 file."""
    gr = get_nx_instrument(gr)
    return get_nx_group(gr, 'source', 'NXsource')

def get_nx_default(f):
    """Get the default nexus group from a nexus hdf5 instance."""
    entry = get_nx_entry(f)
    if entry is None:
        return None
    if 'default' in entry.attrs:
        default = entry.attrs['default']
        if default in entry:
            return entry[default]
    elif 'default' in f:
        default = f.attrs['default']
        if default in f:
            return f[default]
    return None

def get_nx_signal(gr):
    """Get the signal nexus dset from a nexus group."""
    if gr is None:
        return None
    if "NX_class" in gr.attrs and not gr.attrs["NX_class"] == "NXdata":
        gr = get_nx_default(gr)
    if gr is None:
        return None
    if 'signal' in gr.attrs:
        signal = gr.attrs['signal']
        if signal in gr:
            return gr[signal]
    return None

def get_nx_signal_errors(gr):
    """Get the signal errors nexus dset from a nexus group."""
    if gr is None:
        return None
    signal = get_nx_signal(gr)
    if signal is None:
        return None
    error_name = signal.name + '_errors'
    if error_name in gr:
        return gr[error_name]
    return None

def get_nx_axes(gr):
    """Get a list of the axes nexus dsets from a nexus group."""
    if gr is None:
        return []
    if "NX_class" in gr.attrs and not gr.attrs["NX_class"] == "NXdata":
        gr = get_nx_default(gr)
    if gr is None:
        return []
    axes = []
    if 'axes' in gr.attrs:
        axes_names = gr.attrs['axes']
        for ax in axes_names:
            if ax in gr and isinstance(gr[ax], h5.Dataset):
                axes.append(gr[ax])
            else:
                axes.append(None)
    return axes
        
def get_nx_energy(f):
    """Attempt to get the energy from the nxmonochromator group in a nexus hdf5 file"""
    entry = get_nx_entry(f)
    if entry is None:
        return None
    monochromator = get_nx_monochromator(entry)
    if monochromator is None:
        return None
    if 'energy' in monochromator:
        return monochromator['energy'][()]
    elif 'wavelength' in monochromator:
        wavelength = monochromator['wavelength'][()]
        return 12.398 / wavelength  # Convert wavelength to energy in keV
# If no energy or wavelength is found, return None
    return None    

def get_instrument_name(gr):
    """Get the instrument name from a nexus file and return it as a string."""
    instrument = get_nx_instrument(gr)
    if instrument is not None and 'name' in instrument:
        name =  instrument['name'][()]
        # Check if the name is a bytes object and decode it
        if isinstance(name, bytes):
            return name.decode('utf-8')
        return name
    return None

def get_source_name(gr):
    """Get the source name from a nexus file and return it as a string."""
    source = get_nx_source(gr)
    if source is not None and 'name' in source:
        name = source['name'][()]
        # Check if the name is a bytes object and decode it
        if isinstance(name, bytes):
            return name.decode('utf-8')
        return name
    return None

def get_translations_from_nx_transformations(gr):
    """
    Get the translations from the nxtransformations group in a nexus hdf5 file.
    Use the 'transformation_type' attribute to identify translations, and the
    'vector' attribute to identify the axis.
    Returns a dictionary with keys 'x', 'y', 'z' and values as the corresponding datasets.
    """
    transformations = get_nx_transformations(gr)
    if transformations is not None:
        translations = {}
        for name, dset in transformations.items():
            if hasattr(dset, 'attrs') and "transformation_type" in dset.attrs:
                v = dset.attrs.get("vector",None)
                if v is None:
                    continue
                name = [["x","y","z"][i] for i, comp in enumerate(v) if comp == 1][0]
                translations[name] = dset[:]
    return translations

if __name__ == "__main__":
    pass


# -*- coding: utf-8 -*-
"""
plaid - plaid looks at integrated data
F.H. Gjørup 2025-2026
Aarhus University, Denmark
MAX IV Laboratory, Lund University, Sweden

This module provides I/O functions for reading h5py files and exporting data.

FOR USER-DEFINED FILE FORMATS:
To define custom dataset paths for loading data from HDF5 files,
edit the USER_FILE_PARSER.py file in the plaid directory.

"""
import os
import h5py as h5
import numpy as np
from PyQt6.QtCore import pyqtSignal, QObject, QThread, pyqtSlot
from nexus import *
from dialogs import H5Dialog
try:
    from USER_FILE_PARSER import USER_FILE_PARSER
except Exception as e:
    # If the USER_FILE_PARSER.py file does not exist, copy the template
    if not os.path.exists(os.path.join(os.path.dirname(__file__), 'USER_FILE_PARSER.py')):
        with open(os.path.join(os.path.dirname(__file__), 'ufp_temp.py'), 'r') as src:
            content = src.readlines()
        with open(os.path.join(os.path.dirname(__file__), 'USER_FILE_PARSER.py'), 'w') as dst:
            dst.writelines(content[3:])  # Skip the first 3 lines (the comments)
    # if the file exists but cannot be imported, print the error
    else:
        print(f"Error importing USER_FILE_PARSER.py: {e}")

    USER_FILE_PARSER = {"I": None,
                        "I_error": None,
                        "tth": None,
                        "q": None,
                        "energy": None,
                        "wavelength": None,
                        "I0": None,
                        "instrument_name": None,
                        "source_name": None,
                        "map_shape": None,
                        "map_indices": None,
                        }

class Iter_H5Dataset(h5.File):
    """
    A convenient HDF5 file class for iterative data access.
    Inherits from h5py.File.
    """
    def __init__(self, name, dataset_path, **kwargs):
        super().__init__(name, **kwargs)
        self.chunk_size = None
        self.n_chunks = None
        self.dset = self[dataset_path]
        self._set_chunk_size()

    def _set_chunk_size(self, target_mem=4):
        """Attempt to determine a suitable chunk size for a dataset, using
        a target memory size. If the dataset is chunked, use its chunk size as basis.
        Parameters:
            target_mem (int): Target memory size in MB for each chunk.
        """
        single_chunk_size = self.dset.chunks if self.dset.chunks else (1,*self.dset.shape[1:])
        # estimate the memory size of a single chunk
        estimated_chunk_mem = np.prod(single_chunk_size) * self.dset.dtype.itemsize
        # set chunk size to keep memory usage around 4 MB
        target_mem = target_mem * 1024 * 1024  # Convert MB to bytes
        n = max(1, target_mem // estimated_chunk_mem)
        self.chunk_size = n * single_chunk_size[0]
        self.n_chunks = (self.dset.shape[0] + self.chunk_size - 1) // self.chunk_size

    def iter_read(self, chunk_size=None):
        """
        Generator to read a dataset in chunks.
        
        Parameters:
            chunk_size (int): Number of elements to read per chunk.
        """
        if chunk_size is None:
            chunk_size = self.chunk_size
        total_size = self.dset.shape[0]
        for start in range(0, total_size, chunk_size):
            end = min(start + chunk_size, total_size)
            yield self.dset[start:end]

class ReadWorker(QObject):
    """
    A simple QObject worker that runs a callable in a QThread for reading HDF5 
    datasets. The Worker instance owns the QThread during execution and will clean up
    the thread when finished.
    """
    sigStarted = pyqtSignal()
    sigFinished = pyqtSignal(bool, object)  # success(bool), result or exception
    sigError = pyqtSignal(object)
    sigProgress = pyqtSignal(int)  # progress in 0-10000

    def __init__(self, parent=None):
        super().__init__(parent)
        self.fname = None
        self.dataset_path = None
        self._thread = None
        self.success = False
        self.cancelled = False

    @pyqtSlot()
    def _run(self):
        """Internal slot that executes the callable."""
        self.sigStarted.emit()
        try:
            result = self.read_iter(self.fname, self.dataset_path)
            self.sigFinished.emit(True, result)
            self.success = True
        except Exception as e:
            # emit error signals and finished with failure
            self.sigError.emit(e)
            self.sigFinished.emit(False, e)
            self.success = False

    def start(self, fname, dataset_path):
        """Start the worker in a new QThread."""
        if self._thread is not None and self._thread.isRunning():
            raise RuntimeError('Worker already running')
        self.fname = fname
        self.dataset_path = dataset_path
        self.cancelled = False
        self._thread = QThread()
        # move self to thread and start
        #self.moveToThread(self._thread)
        self._thread.started.connect(self._run)
        # ensure cleanup when finished
        self.sigFinished.connect(lambda *_: self._cleanup())
        self._thread.start()

    def read_iter(self,fname, dataset_path):
        """Utility function to read a dataset iteratively using Iter_H5Dataset."""
        import time
        with Iter_H5Dataset(fname, dataset_path) as f:
            data = np.empty_like(f.dset)
            for i,chunk in enumerate(f.iter_read()):
                start = i * f.chunk_size
                end = start + chunk.shape[0]
                data[start:end] = chunk
                self.sigProgress.emit(int(((i+1)*1e4)//f.n_chunks))
                if self.cancelled:
                    break
            return data

    def _cleanup(self):
        try:
            if self._thread is not None:
                self._thread.quit()
                self._thread.wait()
                self._thread = None
        except Exception:
            pass
        
def read_from_dict(f, file_dict):
    """Read datasets from an HDF5 file based on a provided file dictionary."""
    data = {}
    for key, path in file_dict.items():
        if path is not None and path in f:

            data[key] = f[path][()]
        else:
            data[key] = None
    return data

def _determine_file_dict(f, parent=None):
    """Initialize the file dictionary with keys for various data types."""
    file_dict = {"I": None,
                "I_error": None,
                "tth": None,
                "q": None,
                "energy": None,
                "wavelength": None,
                "I0": None,
                "instrument_name": None,
                "source_name": None,
                "map_shape": None,
                "map_indices": None,
                }
    if 'entry/data1d' in f:
        # old (DanMAX) nxazint HDF5 file
        data_group = f['entry/data1d']
        file_dict.update({"I": get_h5_dset_path(data_group,'I'),
                        "I_error": get_h5_dset_path(data_group,'I_error'),
                        "tth": get_h5_dset_path(data_group,'2th'),
                        "q": get_h5_dset_path(data_group,'q'),
                        })
        return file_dict
    elif 'entry/dataxrd1d' in f:
        # DanMAX map HDF5 file
        data_group = f['entry/dataxrd1d']
        file_dict.update({"I": get_h5_dset_path(data_group,'xrd'),
                        "I_error": get_h5_dset_path(data_group,'xrd_error'),
                        "tth": get_h5_dset_path(data_group,'tth'),
                        "q": get_h5_dset_path(data_group,'q'),
                        })
        return file_dict
    elif USER_FILE_PARSER.get("I", None) is not None:
        # Use the user-defined file parser dictionary
        for key, path in USER_FILE_PARSER.items():
            if path is not None and path in f:
                file_dict[key] = path
        # ensure that at least "I" and "tth" or "q" are defined
        if file_dict["I"] is None or (file_dict["tth"] is None and file_dict["q"] is None):
            file_dict = {key:None for key in file_dict.keys()}  
            pass
        else:
            return file_dict
        
    # Attempt to load using the H5Dialog if no specific structure is found
    _selected_dict = _load_dialog(f, parent=parent)
    if _selected_dict is not None:
        file_dict.update(_selected_dict)
    return file_dict

def _load_dialog(f, parent=None):
    """
    Load azimuthal integration data from an h5 file dialog.  
    This function is used as a last resort if no other load function is found.
    """
    dialog = H5Dialog(parent, f)
    if not dialog.exec_1d_2d_pair():
        return None

    selected = dialog.get_selected_items() # list of tuples with (alias, full_path, shape)
    axis = [item for item in selected if not "×" in item[2]][0] 
    signal = [item for item in selected if "×" in item[2]][0]
    # Check if the shape of the axis and signal match
    if not axis[2] in signal[2].split("×")[1]:
        print(f"Error: The shape of the axis {axis[2]} does not match the shape of the signal {signal[2]}.")
        return None
    # attempt to guess if the axis is q or 2theta
    is_q = 'q' in axis[0].lower() or 'q' in f[axis[1]].attrs.get('long_name', '').lower()
    file_dict = {"I": signal[1],
                 "tth": axis[1] if not is_q else None,
                 "q": axis[1] if is_q else None,
                    }
    return file_dict

def load_file(fname, parent=None):
    """
    Load azimuthal integration data from a nexus or generic HDF5 file,
    EXCLUDING the intensity data (and error). Return a dictionary with
    dataset paths for I and I_error and metadata.
    
    data_dict = {"I": None,
                "I_error": None,
                "tth": None,
                "q": None,
                "energy": None,
                "wavelength": None,
                "I0": None,
                "instrument_name": None,
                "source_name": None,
                "map_shape": None,
                "map_indices": None,
                }
    """
    with h5.File(fname, 'r') as f:
        entry = get_nx_entry(f,definition="NXazint1d",allow_subentry=True)
        if entry is not None:
            default = get_nx_default(entry)
            signal = get_nx_signal(default)
            signal_errors = get_nx_signal_errors(default)
            axis = get_nx_axes(default)[-1] # Get the last axis, which is usually the radial axis
            is_Q = 'q' in axis.attrs['long_name'].lower() if 'long_name' in axis.attrs else False
            monochromator = get_nx_monochromator(entry)
            instrument = get_nx_instrument(entry)
            source = get_nx_source(entry)
            monitor = get_nx_monitor(entry)
            data_dict = {"I": signal.name,
                        "I_error": signal_errors.name if signal_errors is not None else None,
                        "tth": axis[:] if not is_Q else None,
                        "q": axis[:] if is_Q else None,
                        "energy": get_nx_energy(monochromator),
                        "wavelength": None,
                        "I0": monitor["data"][:] if monitor is not None else None,
                        "instrument_name": get_instrument_name(instrument),
                        "source_name": get_source_name(source),
                        }

        elif 'entry/dataxrd1d/xrd' in f:
            # DanMAX map HDF5 file
            data_group = f['entry/dataxrd1d']
            data_dict = {"I": data_group['xrd'].name,
                         "I_error": data_group['xrd_error'].name if 'xrd_error' in data_group else None,
                         "tth": data_group['tth'][()] if 'tth' in data_group else None,
                         "q": data_group['q'][()] if 'q' in data_group else None,
                         "energy": f['/entry/measurement/Emax'][()] if '/entry/measurement/Emax' in f else None,
                         "wavelength": None,
                         "I0": None,
                         "instrument_name": None,
                         "source_name": None,
                         "map_shape": data_group['xrd'].shape[1:],
                         "map_indices": list(range(np.prod(data_group['xrd'].shape[1:]))),
                         }
        else:
            file_dict = _determine_file_dict(f,parent=parent)
            data_dict = {"I": file_dict.pop("I"),
                         "I_error":file_dict.pop("I_error"),
                         }

            data_dict.update(read_from_dict(f, file_dict))
    return data_dict

def export_xy(fname, x, y, y_e=None, kwargs={}):
    """
    Export the azimuthal integration data to a text file.  
    kwargs are passed to np.savetxt.
    Parameters:
        fname (str): The file name to save the data to.
        x (np.ndarray): The x-axis data (e.g., tth or q).
        y (np.ndarray): The intensity data.
        y_e (np.ndarray, optional): The intensity error data.
        kwargs (dict, optional): Additional keyword arguments for np.savetxt.
    Returns:
        bool: True if the export was successful.
    """
    if y_e is None:
        np.savetxt(fname, np.column_stack((x, y)),comments='#',**kwargs)
    else:
        np.savetxt(fname, np.column_stack((x, y, y_e)),comments='#',**kwargs)
    return True



# -*- coding: utf-8 -*-
"""
plaid - plaid looks at integrated data
F.H. Gjørup 2025-2026
Aarhus University, Denmark
MAX IV Laboratory, Lund University, Sweden

This module provides functions for miscellaneous calculations related to diffraction data,
including conversions between q and 2theta.
"""
import numpy as np

HC = 12.3984193  # Planck constant times speed of light in keV·Å

def q_to_tth(q, E):
    """Convert q to 2theta."""
    # Convert 2theta to radians
    wavelength = HC / E
    tth = 2 * np.degrees(np.arcsin(q * wavelength / (4 * np.pi)))
    return tth

def tth_to_q(tth, E):
    """Convert 2theta to q."""
    # Convert 2theta to radians
    wavelength = HC / E
    q = (4 * np.pi / wavelength) * np.sin(np.radians(tth) / 2)
    return q

def d_to_q(d):
    """Convert d-spacing to q."""
    return 2 * np.pi / d

def q_to_d(q):
    """Convert q to d-spacing."""
    return 2 * np.pi / q

def d_to_tth(d, E):
    """Convert d-spacing to 2theta."""
    wavelength = HC / E
    tth = 2 * np.degrees(np.arcsin(wavelength / (2 * d)))
    return tth

def tth_to_d(tth, E):
    """Convert 2theta to d-spacing."""
    wavelength = HC / E
    d = wavelength / (2 * np.sin(np.radians(tth) / 2))
    return d


def get_divisors(x):
    """Get all divisors of an integer x."""
    divisors = []
    for i in range(1,int(x**0.5)+1):
        if x%i == 0:
            divisors.append(i)
            divisors.append(x//i)
    return sorted(list(divisors))

def get_map_shape_and_indices(y,x):
    """Get pixel indices and map shape from absolute y (fast) and  x (slow) positions."""
    def guessRes(x,decimals=3):
        """Guess the resolution based on the median of the absolute steps"""
        dx = np.abs(np.diff(x))
        return np.round(np.median(dx[dx>0.0001]),decimals)

    x_res = guessRes(x)
    y_res = guessRes(y)
    x_index = np.round((x-x.min())/x_res).astype(int)
    y_index = np.round((y-y.min())/y_res).astype(int)
    # guess the map shape from the indices
    map_shape = (x_index.max()+1,y_index.max()+1)

    pixel_indices = np.arange(np.prod(map_shape)).reshape(map_shape)
    pixel_indices = list(pixel_indices[x_index,y_index])

    return map_shape, pixel_indices

def average_blocks(arr, reduction_factor=2, axes=(0,)):
    """
    Reduce a numpy array along multiple axes by averaging non-overlapping blocks of the same size.
    If the size along any axis is not divisible by reduction_factor, the last entries are dropped.

    Parameters
    ----------
    arr : np.ndarray
        Input array.
    reduction_factor : int
        Number of adjacent entries to average along each axis.
    axes : int or tuple of int
        Axes along which to average.

    Returns
    -------
    np.ndarray
        Reduced array.
    """
    if reduction_factor is None or reduction_factor <= 1:
        return arr
    arr = np.asarray(arr)
    if not hasattr(axes, '__iter__'):
        axes = (axes,)
    for axis in axes:
        if axis < 0 or axis >= arr.ndim:
            raise ValueError(f"Axis {axis} out of bounds for array of dimension {arr.ndim}")
    for axis in sorted(axes):
        n = arr.shape[axis]
        rf = min(reduction_factor, n)
        trimmed = n - (n % rf)
        if trimmed != n:
            slc = [slice(None)] * arr.ndim
            slc[axis] = slice(0, trimmed)
            arr = arr[tuple(slc)]
        shape = list(arr.shape)
        new_shape = shape[:axis] + [shape[axis] // rf, rf] + shape[axis+1:]
        arr = arr.reshape(new_shape)
        arr = arr.mean(axis=axis+1)
    return arr

if __name__ == "__main__":  
    pass


# -*- coding: utf-8 -*-
"""
plaid - plaid looks at integrated data
F.H. Gjørup 2025-2026
Aarhus University, Denmark
MAX IV Laboratory, Lund University, Sweden

This module provides a class to hold azimuthal integration data and perform various operations on it,
including loading data from HDF5 files, converting between q and 2theta, and normalizing intensity data.

"""
import numpy as np
from PyQt6.QtWidgets import  QInputDialog, QMessageBox
import h5py as h5
from plaid.nexus import (get_nx_monitor, get_nx_sample, get_nx_transformations, 
                         get_translations_from_nx_transformations)
from plaid.misc import q_to_tth, tth_to_q, get_map_shape_and_indices, average_blocks
from plaid.io import export_xy

class AzintData():
    """
    A class to hold azimuthal integration data.
    Parameters:
    - parent: The parent widget, usually the main window.
    - fnames: A list of file names to load the azimuthal integration data from.
    Attributes:
    - x: The radial axis data (2theta or q).
    - I: The intensity data.
    - is_q: A boolean indicating if the radial axis is in q or 2theta.
    - E: The energy data, if available.
    - I0: The I0 data, if available.
    - shape: The shape of the intensity data.
    - instrument_name: The name of the instrument, if available.
    - source_name: The name of the source, if available.
    """

    def __init__(self, parent=None,fnames=None):
        self.parent = parent
        if isinstance(fnames, str):
            fnames = [fnames]
        self.fnames = fnames
        self.x = None
        self.I = None
        self.I_error = None
        #self.y_avg = None
        self.is_q = False
        self.E = None
        self.I0 = None
        self.shape = None  # Shape of the intensity data
        self._shapes = []  # Shapes of individual files loaded
        self.reduction_factor = 1 # Reduction factor applied to the data (compounded if multiple reductions are applied)
        self.instrument_name = None  # Name of the instrument, if available
        self.source_name = None  # Name of the source, if available
        self._load_func = None
        self.map_shape = None  # Shape of the loaded data files used for mapping (PLACEHOLDER)
        self.map_indices = None  # Indices of the loaded data files used for mapping (PLACEHOLDER)

        self.y_bgr = None  # Background intensity data

        #self.aux_data = {} # {alias: np.array}

    def set_secondary_data(self, data_dict):
        """Set the "secondary" azimuthal integration data from a dictionary."""
        is_q = data_dict["q"] is not None
        self.is_q = is_q
        self.x = data_dict["q"] if is_q else data_dict["tth"]
        self.E = data_dict.get("energy", None)
        self.instrument_name = data_dict.get("instrument_name", None)
        self.source_name = data_dict.get("source_name", None)
        self.map_shape = data_dict.get("map_shape", None)
        self.map_indices = data_dict.get("map_indices", None)
    
    def load_I0_from_nxmonitor(self):
        """
        Load the I0 data from a nxmonitor dataset in the HDF5 file(s).
        All files in self.fnames are expected to have a nxmonitor dataset,
        otherwise, it returns None.
        """
        I0 = np.array([])
        for fname in self.fnames:
            with h5.File(fname, 'r') as f:
                monitor = get_nx_monitor(f)
                if monitor is None or 'data' not in monitor:
                    I0_ = None
                else:
                    I0_ = monitor['data'][:]
                if I0_ is None:
                    self.I0 = None
                    return False
                I0 = np.append(I0, I0_) if I0.size else I0_
        self.I0 = I0
        return True
    
    def load_map_shape_and_indices(self):
        """
        Load the map shape and pixel indices from an nxtransformations group
        in the HDF5 file(s). All files in self.fnames are expected to have
        a nxtransformations group, otherwise, it returns None.
        """
        x,y = np.array([]), np.array([])
        for fname in self.fnames:
            with h5.File(fname, 'r') as f:
                sample = get_nx_sample(f)
                if sample is None:
                    self.map_shape = None
                    self.pixel_indices = None
                    return False
                transformations = get_nx_transformations(sample)
                if transformations is None:
                    self.map_shape = None
                    self.pixel_indices = None
                    return False
                translations = get_translations_from_nx_transformations(transformations)
                if translations is None or "x" not in translations or "y" not in translations:
                    self.map_shape = None
                    self.pixel_indices = None
                    return False
                x = np.append(x, translations["x"]) if x.size else translations["x"]
                y = np.append(y, translations["y"]) if y.size else translations["y"]
        self.map_shape, self.map_indices = get_map_shape_and_indices(y, x)
        return True

    def reduce_data(self, reduction_factor=2, axes=(0,)):
        """Reduce the azimuthal integration data by averaging non-overlapping blocks."""
        
        if self.I is None:
            return
        self.I = average_blocks(self.I, reduction_factor=reduction_factor, axes=axes)
        if self.I_error is not None:
            self.I_error = average_blocks(self.I_error, reduction_factor=reduction_factor, axes=axes)
        if self.I0 is not None:
            self.I0 = average_blocks(self.I0, reduction_factor=reduction_factor, axes=axes)
        #self.y_avg = self.I.mean(axis=0) if self.I is not None else None
        self.shape = self.I.shape if self.I is not None else None
        self.map_shape, self.map_indices = None, None  # Invalidate map shape and indices after reduction
        self.reduction_factor *= reduction_factor

    def user_E_dialog(self):
        """Prompt the user for the energy value if not available in the file."""
        if self.E is None:
            E, ok = QInputDialog.getDouble(self.parent, "Energy Input", "Enter the energy in keV:", value=35.0, min=1.0, max=200.0)
            if ok:
                self.E = E
                return E
        else:
            return self.E
    
    def get_tth(self):
        """Calculate the 2theta values from the energy and radial axis."""
        if not self.is_q:
            # If the data is already in 2theta, return it directly
            return self.x
        if self.E is None:
            self.user_E_dialog()
        if self.E is None:
            print("Energy not set. Cannot calculate 2theta.")
            return None
        tth = q_to_tth(self.x, self.E)
        return tth
    
    def get_q(self):
        """Calculate the q values from the energy and radial axis."""
        if self.is_q:
            return self.x
        if self.E is None:
            self.user_E_dialog()
        if self.E is None:
            print("Energy not set. Cannot calculate q.")
            return None
        q = tth_to_q(self.x, self.E)
        return q
        
    def get_I(self, index=None, I0_normalized=True, bgr_subtracted=True):
        """
        Get the intensity data at I[index] if index not None, otherwise return I.
        By default, it returns the normalized intensity data by dividing by I0 if I0 is set.
        """
        if self.I is None:
            print("No intensity data loaded.")
            return None
        I0 = 1
        if self.I0 is not None and I0_normalized:
            if self.I0.shape[0] != self.shape[0]:
                print(f"I0 data shape {self.I0.shape} must match the number of frames {self.shape} in the azimuthal integration data.")
                return None
            I0 = self.I0
        if index is not None:
            I = self.I[index, :]  # Get the intensity data for the specified index
            I0 = I0[index] if isinstance(I0, np.ndarray) else I0  # Get the corresponding I0 value
        else:
            I = self.I
        if bgr_subtracted and self.y_bgr is not None:
            I = I - self.y_bgr
        return (I.T / I0).T
    
    def get_average_I(self, I0_normalized=True,bgr_subtracted=True):
        """Get the average intensity data, normalized by I0 if set."""
        if self.I is None:
            print("No intensity data loaded.")
            return None
        I = self.get_I(index=None, I0_normalized=I0_normalized,bgr_subtracted=bgr_subtracted)
        return np.mean(I, axis=0) if I is not None else None

    def get_I_error(self, index=None, I0_normalized=True):
        """
        Get the intensity errors at I_error[index] if index is not None, otherwise return I_error.
        If I0_normalized is True, normalize the intensity errors by I0.
        """
        if self.I_error is None:
            return None
        I0 = 1
        if self.I0 is not None and I0_normalized:
            if self.I0.shape[0] != self.shape[0]:
                print(f"I0 data shape {self.I0.shape} must match the number of frames {self.shape} in the azimuthal integration data.")
                return None
            I0 = self.I0
        if index is not None:
            I_error = self.I_error[index, :]
            I0 = I0[index] if isinstance(I0, np.ndarray) else I0
        else:
            I_error = self.I_error
        return (I_error.T / I0).T if I_error is not None else None

    def get_average_I_error(self, I0_normalized=True):
        """Get the average intensity errors, normalized by I0 if set."""
        if self.I_error is None:
            return None
        I_error = self.get_I_error(index=None, I0_normalized=I0_normalized)
        return np.mean(I_error, axis=0) if I_error is not None else None

    def set_y_bgr(self, y_bgr):
        """Set the background intensity data."""
        if y_bgr is None:
            self.y_bgr = None
            return
        if isinstance(y_bgr, np.ndarray):
            self.y_bgr = y_bgr
        elif isinstance(y_bgr, (list, tuple)):
            self.y_bgr = np.array(y_bgr)
        else:
            print("Background intensity data must be a numpy array or a list/tuple.")
            return
        if self.y_bgr.shape[0] != self.x.shape[0]:
            print(f"Background intensity data shape {self.y_bgr.shape} must match the radial axis shape {self.x.shape}.")
            self.y_bgr = None
            return
        
    def set_I0(self, I0):
        """Set the I0 data."""
        if isinstance(I0, np.ndarray):
            self.I0 = I0
        elif isinstance(I0, (list, tuple)):
            self.I0 = np.array(I0)
        else:
            print("I0 data must be a numpy array or a list/tuple.")
            return
        
        if self.I is None:
            # Don't normalize (yet)
            return
        
        if self.I.shape[0]  != self.I0.shape[0]:
            print(f"I0 data shape {self.I0.shape} must match the number of frames {self.I.shape} in the azimuthal integration data.")
            return
    
    def export_pattern(self, fname, index, is_Q=False, I0_normalized=True, kwargs={}):
        """
        Export the azimuthal integration data at the current index to a text file.  
        If I0_normalized is True, normalize the intensity data by I0.  
        kwargs passed to np.savetxt  
        """
        if self.I is None:
            print("No intensity data loaded.")
            return False
        if is_Q:
            x = self.get_q()
        else:
            x = self.get_tth()
        y = self.get_I(index=index, I0_normalized=I0_normalized)
        y_e = self.get_I_error(index=index, I0_normalized=I0_normalized)
        if x is None or y is None:
            print("Error retrieving data for export.")
            return False
        
        export_xy(fname,x,y,y_e, kwargs)
        return True
    
    def export_average_pattern(self, fname, is_Q=False, I0_normalized=True, bgr_subtracted=True, kwargs={}):
        """
        Export the average azimuthal integration data to a text file.  
        If I0_normalized is True, normalize the intensity data by I0.  
        kwargs passed to np.savetxt  
        """
        if self.I is None:
            print("No intensity data loaded.")
            return False
        if is_Q:
            x = self.get_q()
        else:
            x = self.get_tth()
        y = self.get_average_I(I0_normalized=I0_normalized, bgr_subtracted=bgr_subtracted)
        y_e = self.get_average_I_error(I0_normalized=I0_normalized, bgr_subtracted=bgr_subtracted)
        
        if x is None or y is None:
            print("Error retrieving data for export.")
            return False

        export_xy(fname,x,y,y_e, kwargs)
        return True
    
    def get_info_string(self):
        """Get the instrument (and source) name from the azimuthal integration data."""
        name = ""
        if self.instrument_name is not None:
            name += self.instrument_name
        if self.source_name is not None:
            if name:
                name += " - "
            name += self.source_name
        if self.E is not None:
            if name:
                name += " - "
            name += f"energy: {self.E:.2f} keV"
        if self.I0 is not None:
            if name:
                name += " - "
            name += f"I0 corrected"
        if self.reduction_factor != 1:
            if name:
                name += " - "
            name += f"reduced x{self.reduction_factor}"
        return name

class AuxData:
    """A class to hold auxiliary data for azimuthal integration."""
    def __init__(self,parent=None):
        self._parent = parent
        self.I0 = None
        self._E = None

    def set_energy(self, E):
        """Set energy"""
        self._E = E
        
    def get_energy(self):
        """Get energy"""
        return self._E

    def set_I0(self, I0):
        """Set I0"""
        if isinstance(I0, (tuple, list)):
            I0 = np.array(I0)

        # check if the I0 data are close to unity
        # otherwise, normalize it and print a warning
        if I0.min() <= 0 or I0.max() < 0.5 or I0.max() > 2:
            if self._parent:
                message = ("Warning: I0 data should be close to unity and >0. Normalizing it.\n"
                            f" I0 [{I0.min():.2e}, {I0.max():.2e}] normalized to [{I0.min()/I0.max():.2f}, 1.00]")
                QMessageBox.warning(self._parent, "I0 Data Warning", message)
            else:
                print("Warning: I0 data should be close to unity and >0. Normalizing it.")
                print(f"I0 [{I0.min():.2e}, {I0.max():.2e}] normalized to [{I0.min()/I0.max():.2f}, 1.00]")
            I0 = I0 / np.max(I0)
            I0[I0<=0] = 1  # Set any zero values to 1 to avoid division by zero
        self.I0 = I0

    def add_data(self, key, data):
        """Add data to the AuxData instance."""
        if isinstance(data, (list, tuple)):
            data = np.array(data)
        setattr(self, key, data)

    def get_data(self, key):
        """Get data from the AuxData instance."""
        if isinstance(key, (list, tuple)):
            return [self.get_data(k) for k in key]
        if not hasattr(self, key):
            print(f"Key '{key}' not found in AuxData.")
            return None
        return getattr(self, key, None)
    
    def get_dict(self):
        """Get a dictionary representation of the AuxData instance."""
        return {key: value for key, value in self.__dict__.items() if not key.startswith('_')}
    
    def keys(self):
        """Get the keys of the AuxData instance."""
        return [key for key in self.__dict__.keys() if not key.startswith('_')]
        
    def clear(self):
        """Clear all data in the AuxData instance."""
        self.__dict__.clear()
        self.I0 = None


if __name__ == "__main__":
    pass


### USER FILE PARSER DICTIONARY TEMPLATE ###
### DO NOT EDIT THIS FILE DIRECTLY!      ###
### EDIT 'USER_FILE_PARSER.py' INSTEAD!  ###
# -*- coding: utf-8 -*-
"""
plaid - plaid looks at integrated data
F.H. Gjørup 2025-2026
Aarhus University, Denmark
MAX IV Laboratory, Lund University, Sweden

User-defined HDF5 file parser dictionary.
Use this dictionary to define custom dataset paths for loading data from HDF5 files.
Each key corresponds to a specific data type, and the value should be the dataset path as a string.
Only "I" and "tth" or "q" are required. Other fields can be set to None if not available.

Restart the application after editing this file to apply changes.

Example:
USER_FILE_PARSER = {"I": "entry/data/intensity",
                    "tth": "entry/data/two_theta",
                    }
"""

USER_FILE_PARSER = {"I": None,
                    "I_error": None,
                    "tth": None,
                    "q": None,
                    "energy": None,
                    "wavelength": None,
                    "I0": None,
                    "instrument_name": None,
                    "source_name": None,
                    "map_shape": None,
                    "map_indices": None,
                    }